"""
    The `clear_sky_radiation` module contains all functions related to the
    calculation of (instantaneous) clear sky radation. Most of these
    functions are based upon :footcite:t:`vsuri2004new`.

"""
import numpy as np
import xarray as xr

def extraterrestrial_irradiance_normal(I0, ied):
    r"""
    Computes the extraterrestrial irradiance normal to the solar beam.

    .. math ::

        G_{0} = I_{0} \cdot \varepsilon

    Parameters
    ----------
    I0 : float
        solar constant, 
        :math:`I_{0}`
        [W m\ :sup:`-2`\]
    ied : float
        inverse earth sun distance, 
        :math:`\varepsilon`
        [AU\ :sup:`-1`\]

    Returns
    -------
    G0 : float
        ext rad normal to solar bea, 
        :math:`G_{0}`
        [W m\ :sup:`-2`\]

    """
    return I0*ied


def inverse_earth_sun_distance(day_angle):
    r"""
    Computes the inverse earth sun distance

    .. math ::

        \varepsilon = 1 + 0.03344 \cdot \cos(j^{\prime} - 0.048869)

    Parameters
    ----------
    day_angle : float
        day angle, 
        :math:`j^{\prime}`
        [-]

    Returns
    -------
    ied : float
        inverse earth sun distance, 
        :math:`\varepsilon`
        [AU]

    """
    return 1 + 0.03344 * np.cos(day_angle - 0.048869)


def day_angle(doy):
    r"""

    Computes the day angle. 0 is january 1\ :sup:`st`\,
    2\ :math:`\pi` is december 31\ :sup:`st`\.

    .. math ::

        j^{\prime} = \frac{2 \cdot \pi \cdot j} {365.25}

    Parameters
    ----------
    doy : float
        day of year, 
        :math:`j`
        [-]

    Returns
    -------
    day_angle : float
        day angle, 
        :math:`j^{\prime}`
        [rad]

    """
    return 2 * np.pi * doy / 365.25


def solar_constant():
    r"""
    Returns the solar constant. The solar constant is defined as the
    flux density of solar radiation at the mean distance from Sun
    to Earth. The solar constant is estimated to be 1367 W m\ :sup:`-2`\.

    Parameters
    ----------

    Returns
    -------
    I0 : float
        solar constant, 
        :math:`I_{0}`
        [W m\ :sup:`-2`\]
    """
    return 1367.


def declination(day_angle):
    r"""
    Computes the solar declination. The solar declination is computed according to 
    :footcite:t:`gruter1984radiation`.

    .. math ::

        \delta = \arcsin\left(0.3978 \cdot \sin\left(j^{\prime} - 1.4 + 0.0355 \cdot \sin
              \left(j^{\prime} - 0.0489\right)\right)\right)

    Parameters
    ----------
    day_angle : float
        day angle, 
        :math:`j^{\prime}`
        [rad]

    Returns
    -------
    decl : float
        declination, 
        :math:`\delta`
        [rad]

    """
    return np.arcsin(0.3978 * np.sin(day_angle - 1.4 + 0.0355 * np.sin(day_angle - 0.0489)))


def relative_optical_airmass(p_air_i, p_air_0_i, h0ref):
    r"""
    Computes the relative optical air mass. It is calculated according to :footcite:t:`kasten1989revised`.

    .. math ::

        m = \frac{\frac{p}{p_{0}}}{\sin h_{0}^{ref}+0.50572
         \left(h_{0}^{ref}+6.07995\right)^{-1.6364}}

    Parameters
    ----------
    p_air_i : float
        actual instantaneous air pressure, 
        :math:`p`
        [hPa]
    p_air_0_i : float
        air pressure at sea level, 
        :math:`p_{0}`
        [-]
    h0ref : float
        solar elevation angle corrected for refraction, 
        :math:`h_{0}^{ref}`
        [degrees]

    Returns
    -------
    m : float
        relative optical airmass, 
        :math:`m`
        [-]

    """
    h0ref_rad = h0ref * np.pi/180.

    m = (p_air_i/p_air_0_i)/(np.sin(h0ref_rad) + 0.50572 * (h0ref + 6.07995)**-1.6364)
    
    if isinstance(m, np.ndarray):
        m[h0ref_rad <= 0] = 64
    else:
        m = m.where(h0ref_rad > 0, 64)

    return m


def solar_elevation_angle_refracted(h0):
    r"""
    Computes the solar elevation angle corrected for refraction.

    .. math ::

        h_{0}^{ref} = h_{0} + \Delta h_{0}^{ref}

    where:

    .. math ::

        \Delta h_{0}^{ref}=0.61359 \cdot \frac{0.1594 + 1.123 \cdot h_{0} +
        0.065656 \cdot h_{0}^{2}}{
        1+28.9344 \cdot h_{0} + 277.3971 \cdot h_{0}^{2}}

    Parameters
    ----------
    h0 : float
        solar elevation angle, 
        :math:`h_{0}`
        [degrees]

    Returns
    -------
    h0ref : float
        solar elevation angle corrected for refrection, 
        :math:`h_{0}^{ref}`
        [degrees]
    """
    delta_h0ref = 0.061359 * (0.1594 + 1.123*h0 + 0.065656*h0**2)/(1 + 28.9344*h0 + 277.3971*h0**2)
    h0ref = h0 + delta_h0ref

    return h0ref


def hour_angle(solar_time):
    r"""
    Computes the solar hour angle.

    .. math ::

        T = \frac{\pi}{12}\left(t-12\right)

    Parameters
    ----------
    solar_time : float
        solar_time, 
        :math:`t`
        [hours]

    Returns
    -------
    ha : float
        solar hour angle, 
        :math:`T`
        [rad]

    """
    ha = np.pi / 12 * (solar_time - 12)

    return ha


def solar_elevation_angle(lat, decl, ha):
    r"""
    Computes the solar elevation angle.

    .. math ::

        h_{0}=\arcsin\left(C_{31} \cdot \cos(T) + C_{33}\right)

    where:

    .. math ::

        C_{31} &= \cos(\varphi) \cdot \cos(\delta) \\
        C_{33} &= \sin(\varphi) \cdot \sin(\delta)

    Parameters
    ----------
    lat : float
        latitude, 
        :math:`\varphi`
        [rad]
    decl : float
        declination, 
        :math:`\delta`
        [rad]
    ha : float
        solar hour angle, 
        :math:`T`
        [rad]

    Returns
    -------
    h0 : float
        solar elevation angle, 
        :math:`h_0`
        [degrees]

    """
    C31 = np.cos(lat) * np.cos(decl)
    C33 = np.sin(lat) * np.sin(decl)

    sin_h0 = C31 * np.cos(ha) + C33

    h0_rad = np.arcsin(sin_h0)

    h0 = h0_rad * 180 / np.pi

    if isinstance(h0, xr.DataArray):
        h0 = h0.transpose("time", "y", "x").chunk("auto")

    return h0


def rayleigh_optical_thickness(m):
    r"""
    Computes the Rayleigh optical thickness at airmass :math:`m`. It is calculated according
    to the improved formula by :footcite:t:`kasten1996linke`.

    if :math:`m` > 20:

    .. math ::

        \delta_{R}(m) = \left(6.6296 + 1.7513 \cdot  m - 0.1202 \cdot m^{2} + 0.0065 \cdot m^{3} - 0.00013 \cdot m^{4}\right)^{-1}

    if :math:`m` < 20:

    .. math ::

        \delta_{R}(m) = \left(10.4+0.718 \cdot m\right)^{-1}

    Parameters
    ----------
    m : float
        relative optical airmass, 
        :math:`m`
        [-]

    Returns
    -------
    rotm : float
        Rayleigh optical thickness at airmass m, 
        :math:`\delta_{R}`
        [-]

    """

    rotm = 1 / (10.4 + 0.718 * m)
    if isinstance(rotm, np.ndarray):
        rotm = np.where(m <= 20, 1/(6.6296 + 1.7513*m - 0.1202*m**2 + 0.0065*m**3 - 0.00013*m**4), rotm)
    else:
        rotm_star = 1/(6.6296 + 1.7513*m - 0.1202*m**2 + 0.0065*m**3 - 0.00013*m**4)
        rotm = rotm_star.where(m <= 20, rotm)

    return rotm


def beam_irradiance_normal_clear(G0, Tl2, m, rotm, h0):
    r"""
    Computes the clear sky beam irradiance normal to the solar beam.

    .. math ::

        B_{0c}=G_{0} \cdot \exp\left(-0.8662 \cdot T_{LK} \cdot m \cdot \delta_{R}\right)

    Parameters
    ----------
    G0 : float
        ext rad normal to solar beam, 
        :math:`G_0`
        [W/m2]
    Tl2 : float
        airmass  2 Linke atmospheric turbidity factor, 
        :math:`T_{LK}`
        [-]
    m : float
        relative optical airmass, 
        :math:`m`
        [-]
    rotm : float
        Rayleigh optical thickness at airmass m, 
        :math:`\delta_{R}`
        [-]
    h0 : float
        solar elevation angle, 
        :math:`h_0`
        [degrees]

    Returns
    -------
    B0c : float
        beam irradiance normal to the solar beam, 
        :math:`B_{0c}`
        [W/m2]

    """
    B0c = G0 * np.exp(-0.8662 * Tl2 * m * rotm)
    if isinstance(B0c, np.ndarray):
        B0c[h0 < 0] = 0
    else:
        B0c = B0c.where(h0 >= 0, 0)

    return B0c


def beam_irradiance_horizontal_clear(B0c, h0):
    r"""
    Computes the clear sky beam irradiance on a horizontal surface.

    .. math ::

        B_{hc} = B_{0c} \cdot \sin\left(h_{0}\right)

    Parameters
    ----------
    B0c : float
        beam irradiance normal to the solar beam, 
        :math:`B_{0c}`
        [W/m2]
    h0 : float
        solar elevation angle, 
        :math:`h_0`
        [degrees]

    Returns
    -------
    Bhc : float
        beam irradiance at a horizontal surface, 
        :math:`B_{hc}`
        [W/m2]

    """
    Bhc = B0c * np.sin(h0 * np.pi / 180)
    if isinstance(Bhc, np.ndarray):
        Bhc[h0 < 0] = 0
    else:
        Bhc = Bhc.where(h0 >= 0, 0)

    return Bhc


def linke_turbidity(wv_i, aod550_i, p_air_i, p_air_0_i):
    r"""
    Computes the air mass 2 Linke atmospheric turbidity factor.

    .. math ::

        p_{rel} &= \frac{p}{p_{0}} \\
        T_{LK} &= 3.91 \cdot \tau_{550} \cdot e^{0.689p_{rel}}+0.376 \cdot \ln\left(TCWV\right)+\left(2+0.54 \cdot p_{rel}-0.34 \cdot p_{rel}^{2}\right)

    Parameters
    ----------
    wv_i : float
        total column atmospheric water vapor, 
        :math:`TCWV`
        [kg m-2]
    aod550_i : float
        Aerosol optical depth at 550nm, 
        :math:`aod550`
        [-]
    p_air_i : float
        actual instantaneous air pressure, 
        :math:`p`
        [hPa]
    p_air_0_i : float
        air pressure at sea level, 
        :math:`p_0`
        [-]

    Returns
    -------
    Tl2 : float
        Airmass 2 Linke atmospheric turbidity factor, 
        :math:`T_{LK}`
        [-]

    Examples
    --------
    """
    # prel = p0 / p # Papers mixes p/p0 and p0/p????
    prel = p_air_i / p_air_0_i

    term1 = 3.91 * np.exp(0.689 * prel) * aod550_i
    term2 = 0.376 * np.log(wv_i)

    Tl2 = term1 + term2 + (2 + 0.54 * prel - 0.5 * prel**2 + 0.16 * prel**2)

    return Tl2


def diffuse_irradiance_horizontal_clear(G0, Tl2, h0):
    r"""
    Computes the clear sky beam irradiance on a horizontal surface.

    .. math ::

        D_{hc}=G_{0} \cdot Tn\left(T_{LK}\right) \cdot F_{d}\left(h_{0}\right)

    For the estimation of the transmission function :math:`Tn\left(T_{LK}\right)`
    the following function is used:

    .. math ::

        Tn\left(T_{LK}\right)=-0.015843+0.030543 \cdot T_{LK}+0.0003797 \cdot T_{LK}^{2}

    The solar altitude function :math:`F_{d}\left(h_{0}\right)` is evaluated using the expression:

    .. math ::

        F_{d}\left(h_{0}\right)=A_{1}+A_{2} \cdot \sin (h_{0})+A_{3} \cdot sin^{2}(h_{0})

    with:

    .. math ::

        A_{1}^{\prime} &= 0.26463-0.061581 \cdot T_{LK}+0.0031408 \cdot T_{LK}^{2} \\
        A_{1} &= \frac{0.0022}{Tn\left(T_{LK}\right)} \: \text{if} \: A_{1}^{\prime} \cdot Tn\left(T_{LK}\right)<0.0022 \\
        A_{1} &= A_{1}^{\prime} \: \text{if} \: A_{1}^{\prime} \cdot Tn\left(T_{LK}\right)\geq0.0022 \\
        A_{2} &= 2.04020+0.018945 \cdot T_{LK}-0.011161 \cdot T_{LK}^{2} \\
        A_{3} &= -1.3025+0.039231 \cdot T_{LK}+0.0085079 \cdot T_{LK}^{2}

    Parameters
    ----------
    G0 : float
        ext rad normal to solar beam, 
        :math:`G_0`
        [W/m2]
    Tl2 : float
        Airmass 2 Linke atmospheric turbidity factor, 
        :math:`T_{LK}`
        [-]
    h0 : float
        solar elevation angle, 
        :math:`h_0`
        [degrees]

    Returns
    -------
    Dhc : float
        Diffuse irradiance at a horizontal surface, 
        :math:`D_{hc}`
        [W/m2]

    Examples
    --------
    """
    h0_rad = h0 * np.pi / 180.
    TnTl2 = -0.015843 + 0.030543 * Tl2 + 0.0003797 * Tl2**2

    A1 = 0.26463 - 0.061581 * Tl2 + 0.0031408 * Tl2**2
    if isinstance(A1, xr.DataArray):
        A1 = xr.where((A1 * TnTl2) < 0.0022, 0.0022 / TnTl2, A1)
    else:
        A1 = np.where((A1 * TnTl2) < 0.0022, 0.0022 / TnTl2, A1)
    A2 = 2.04020 + 0.018945 * Tl2 - 0.011161 * Tl2**2
    A3 = -1.3025 + 0.039231 * Tl2 + 0.0085079 * Tl2**2

    FdH0 = A1 + A2 * np.sin(h0_rad) + A3 * np.sin(h0_rad)**2

    Dhc = G0 * TnTl2 * FdH0

    if isinstance(Dhc, np.ndarray):
        Dhc[Dhc < 0] = 0
    else:
        Dhc = Dhc.where(Dhc >= 0, 0)
    
    return Dhc


def ra_clear_horizontal(Bhc, Dhc):
    r"""    
    Computes the clear sky beam irradiance on a horizontal surface.

    .. math ::

        G_{hc}=B_{hc}+D_{hc}

    Parameters
    ----------
    Bhc : float
        beam irradiance at a horizontal surface, 
        :math:`B_{hc}`
        [W/m2]
    Dhc : float
        Diffuse irradiance at a horizontal surface, 
        :math:`D_{hc}`
        [W/m2]
        
    Returns
    -------
    ra_hor_clear_i : float
        Total clear-sky irradiance on a horizontal surface, 
        :math:`G_{hc}`
        [W/m2]

    Examples
    --------
    """
    ra_clear_hor = Bhc + Dhc
    
    return ra_clear_hor



import os
import tqdm
import tempfile
import xarray as xr
from osgeo import gdal, osr
from pywapor.general.logger import log
from pywapor.general.performance import format_bytes
from pywapor.general.processing_functions import save_ds, remove_ds
gdal.UseExceptions()

def calc_slope_or_aspect(ds, var, write_init = True, max_cache_size = 4e9):

    log.add()

    # Get filepath from xr.Dataset.
    fp = ds.encoding.get("source", None)

    if not var in ["slope", "aspect"]:
        raise ValueError("Please set `var` to either `'slope'` or `'aspect'`.")

    # Write data to file if necessary.
    if write_init or isinstance(fp, type(None)):
        fp = os.path.join(tempfile.gettempdir(), f"temp_{var}.nc")
        if os.path.isfile(fp):
            remove_ds(fp)
        while os.path.isfile(fp):
            fp = fp.replace(".nc", "_.nc")
        ds = save_ds(ds, fp, label = f"Generating {var} input file.")

    fp_out = fp.replace(".nc", f"_{var}.tif")

    ds_ = gdal.Open(fp)
    subdss_ = ds_.GetSubDatasets()

    if len(subdss_) > 0:
        fp = [x[0] for x in subdss_ if x[0].split(":")[-1] == "z"][0]
        ds_ = gdal.Open(fp)

    if ds_.RasterCount > 1:
        var_names = [ds_.GetRasterBand(i+1).GetMetadata().get("NETCDF_VARNAME", "unknown") for i in range(ds_.RasterCount)]
        band_select = var_names.find("z") if "z" in var_names else None
        if isinstance(band_select, type(None)):
            log.warning(f"--> Multiple bands found in `{fp}` and none is named `z`, using first band.")
            band_select = 1
    elif ds_.RasterCount == 1:
        band_select = 1
    else:
        raise ValueError(f"No bands found in `{fp}`.")

    band = ds_.GetRasterBand(band_select)
    wkt = ds_.GetProjection()
    src = osr.SpatialReference()
    src.ImportFromWkt(wkt)
    auth = src.GetAuthorityName(None)
    code = src.GetAuthorityCode(None)

    if f"{auth}:{code}" == "EPSG:4326":
        scale = band.GetScale()
        scale_ = {True: 1, False: scale}[isinstance(scale, type(None))] # NOTE in case there is no scale-factor, we set it to 1.
        scale_factor = int(111120 / scale_)
    else:
        scale_factor = 1
        log.warning("--> Dataset has different projection than `EPSG:4326`,\
                    slope scaling might be incorrect if CRS unit is not meters.")

    # Create waitbar.
    waitbar = tqdm.tqdm(position = 0, total = 100, bar_format='{l_bar}{bar}|', delay = 5)

    # Define callback function for waitbar progress.
    def _callback_func(info, *args):
        waitbar.update(info * 100 - waitbar.n)

    options = gdal.DEMProcessingOptions(
        slopeFormat = "degree",
        scale = scale_factor,
        computeEdges = True,
        zeroForFlat = True,
        callback = _callback_func,
    )

    f_size = lambda nbytes: "{0:.2f}{1}".format(*format_bytes(nbytes))

    current_cache_size = gdal.GetCacheMax()
    if "NETCDF:" in fp:
        fp_ = fp.split(":")[1].replace('"', '')
    else:
        fp_ = fp
    
    if os.path.isfile(fp_):
        filesize = os.path.getsize(fp_)
    else:
        filesize = 0.5 * current_cache_size

    if current_cache_size < 0.95 * filesize:
        # NOTE The maximum practical value on 32 bit OS is between 2 and 4 GB. It is the responsibility 
        # of the user to set a consistent value. See: https://gdal.org/user/configoptions.html
        new_cache_size = min(max_cache_size, 2.0 * filesize)
        log.warning(f"--> Input file ({f_size(filesize)}) is larger than current maximum allowed GDAL cache-size ({f_size(current_cache_size)}), will try increasing it to {f_size(new_cache_size)}.")
        gdal.SetCacheMax(int(new_cache_size))
        log.info(f"--> Maximum cache size is {f_size(gdal.GetCacheMax())}.")
    else:
        new_cache_size = None

    ds__ = gdal.DEMProcessing(fp_out, fp, var, options = options)
    ds__.FlushCache()
    ds__ = None

    if not isinstance(new_cache_size, type(None)):
        gdal.SetCacheMax(int(current_cache_size))
        log.info(f"--> Maximum cache size is reset to {f_size(gdal.GetCacheMax())}.")

    ds_new = xr.open_dataset(fp_out).isel(band=0, drop=True).drop_vars(["x", "y"])
    ds_new = ds_new.rename({"band_data": var})
    ds_new.attrs = {}
    for var in ds_new.data_vars:
        ds_new[var].attrs = {}
    ds[var] = ds_new[var]

    log.sub()

    return ds

if __name__ == "__main__":

    import xarray as xr

    # ds = xr.open_dataset(r"/Users/hmcoerver/Local/aspect_slope_test/GLO30_stitched.nc", decode_coords = "all")

    var = "slope"
    write_init = False
    max_cache_size = 4e9


import datetime
import os
import json
import pywapor
import numpy as np
from functools import partial
import pywapor.collect.accounts as accounts
from shapely.geometry.polygon import Polygon
from pywapor.general.logger import log
from shapely.geometry import shape
from pywapor.collect.protocol.projections import get_crss
import pywapor.collect.protocol.opendap as opendap
import xarray as xr
from pywapor.general.processing_functions import open_ds, remove_ds, save_ds
from pywapor.general import bitmasks
import pandas as pd
import warnings
import copy

def fn_func(product_name, tile):
    """Returns a client-side filename at which to store data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        Filename.
    """
    fn = f"{product_name}_h{tile[0]:02d}v{tile[1]:02d}.nc"
    return fn

def url_func(product_name, tile):
    """Returns a url at which to collect MODIS data.

    Parameters
    ----------
    product_name : str
        Name of the product to download.
    tile : str
        Name of the server-side tile to download.

    Returns
    -------
    str
        The url.
    """
    url = f"https://opendap.cr.usgs.gov/opendap/hyrax/{product_name}/h{tile[0]:02d}v{tile[1]:02d}.ncml.nc4?"
    return url

def tiles_intersect(latlim, lonlim):
    """Creates a list of server-side filenames for tiles that intersect with `latlim` and
    `lonlim` for the selected product. 

    Parameters
    ----------
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.

    Returns
    -------
    list
        Server-side filenames for tiles.
    """
    with open(os.path.join(pywapor.collect.__path__[0], "product/MODIS_tiles.geojson")) as f:
        features = json.load(f)["features"]
    aoi = Polygon.from_bounds(lonlim[0], latlim[0], lonlim[1], latlim[1])
    tiles = list()
    for feature in features:
        shp = shape(feature["geometry"])
        tile = feature["properties"]["Name"]
        if shp.intersects(aoi):
            h, v = tile.split(" ")
            htile = int(h.split(":")[-1])
            vtile = int(v.split(":")[-1])
            tiles.append((htile, vtile))
    return tiles

def shortwave_r0(ds, *args):
    ds["r0"] = 0.3 * ds["white_r0"] + 0.7 * ds["black_r0"]
    ds = ds.drop_vars(["white_r0", "black_r0"])
    return ds

def expand_time_dim(ds, *args):
    """MODIS lst data comes with a variable specifying the acquisition decimal time per pixel, This function
    expands the "date" dimension of the data with "time", i.e. afterwards each temporal-slice in the dataset
    contains data at one specific datetime.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset, should have `lst_hour` variable.

    Returns
    -------
    xr.Dataset
        Expanded dataset.
    """

    groups = ds.groupby(ds.lst_hour, squeeze = True)

    def _expand_hour_dim(x):
        hour = np.timedelta64(int(np.nanmedian(x.lst_hour.values) * 3600 * 1e9), "ns")
        x = x.assign_coords({"hour": hour}).expand_dims("hour")
        return x

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Slicing with an out-of-order index")

        ds_expand = groups.map(_expand_hour_dim)

        ds_expand = ds_expand.stack({"datetime": ("hour","time")})

        new_coords = [time + hour for time, hour in zip(ds_expand.time.values, ds_expand.hour.values)]
        
        try: # new versions of xarray require to drop all dimensions of a multi-index
            ds_expand = ds_expand.drop_vars(["datetime", "hour", "time"])
        except ValueError: # old versions throw an error when trying to drop sub-dimensions of a multiindex.
            ds_expand = ds_expand.drop_vars(["datetime"])
        
        ds_expand = ds_expand.assign_coords({"datetime": new_coords}).rename({"datetime": "time"}).sortby("time")
        ds_expand = ds_expand.drop_vars(["lst_hour"])
        ds_expand = ds_expand.transpose("time", "y", "x")
        ds_expand = ds_expand.dropna("time", how="all")

    return ds_expand

def mask_bitwise_qa(ds, var, masker = "lst_qa", 
                product_name = "MOD11A1.061", flags = ["good_qa"]):
    """Mask MODIS data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Variable in `ds` to mask.
    masker : str, optional
        Variable in `ds` to use for masking, by default "lst_qa".
    product_name : str, optional
        Name of the product, by default "MOD11A1.061".
    flags : list, optional
        Which flags not to mask, by default ["good_qa"].

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """

    new_data = ds[var]

    flag_bits = bitmasks.MODIS_qa_translator(product_name)
    mask = bitmasks.get_mask(ds[masker].astype("uint8"), flags, flag_bits)
    new_data = ds[var].where(mask, np.nan)
    ds = ds.drop_vars([masker])

    ds[var] = new_data

    return ds

def mask_qa(ds, var, masker = ("ndvi_qa", 1.0)):
    """Mask MODIS data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data
    var : str
        Variable name in `ds` to be masked.
    masker : tuple, optional
        Variable in `ds` to use for masking, second value defines which value in mask to 
        use as valid data, by default ("ndvi_qa", 1.0).

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """

    new_data = ds[var].where(ds[masker[0]] != masker[1], np.nan)
    ds = ds.drop_vars(masker[0])

    ds[var] = new_data

    return ds

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {

        "MOD13Q1.061": {
                    "_250m_16_days_NDVI": [("time", "YDim", "XDim"), "ndvi"],
                    "_250m_16_days_pixel_reliability": [("time", "YDim", "XDim"), "ndvi_qa"],
                    "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MYD13Q1.061": {
                    "_250m_16_days_NDVI": [("time", "YDim", "XDim"), "ndvi"],
                    "_250m_16_days_pixel_reliability": [("time", "YDim", "XDim"), "ndvi_qa"],
                    "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MOD11A1.061": {
                    "LST_Day_1km": [("time", "YDim", "XDim"), "lst"],
                    "Day_view_time": [("time", "YDim", "XDim"), "lst_hour"],
                    "QC_Day": [("time", "YDim", "XDim"), "lst_qa"],
                    "MODIS_Grid_Daily_1km_LST_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MYD11A1.061": {
                    "LST_Day_1km": [("time", "YDim", "XDim"), "lst"],
                    "Day_view_time": [("time", "YDim", "XDim"), "lst_hour"],
                    "QC_Day": [("time", "YDim", "XDim"), "lst_qa"],
                    "MODIS_Grid_Daily_1km_LST_eos_cf_projection": [(), "spatial_ref"],
                        },
        "MCD43A3.061": {
                    "Albedo_WSA_shortwave": [("time", "YDim", "XDim"), "white_r0"],
                    "Albedo_BSA_shortwave": [("time", "YDim", "XDim"), "black_r0"],
                    "BRDF_Albedo_Band_Mandatory_Quality_shortwave": [("time", "YDim", "XDim"), "r0_qa"],
                    "MOD_Grid_BRDF_eos_cf_projection": [(), "spatial_ref"]
        }
    }

    req_dl_vars = {
        "MOD13Q1.061": {
            "ndvi": ["_250m_16_days_NDVI", "_250m_16_days_pixel_reliability", "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection"],
        },
        "MYD13Q1.061": {
            "ndvi": ["_250m_16_days_NDVI", "_250m_16_days_pixel_reliability", "MODIS_Grid_16DAY_250m_500m_VI_eos_cf_projection"],
        },
        "MOD11A1.061": {
            "lst": ["LST_Day_1km", "Day_view_time", "QC_Day", "MODIS_Grid_Daily_1km_LST_eos_cf_projection"],
        },
        "MYD11A1.061": {
            "lst": ["LST_Day_1km", "Day_view_time", "QC_Day", "MODIS_Grid_Daily_1km_LST_eos_cf_projection"],
        },
        "MCD43A3.061": {
            "r0": ["Albedo_WSA_shortwave", "Albedo_BSA_shortwave", "BRDF_Albedo_Band_Mandatory_Quality_shortwave", "MOD_Grid_BRDF_eos_cf_projection"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = None):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "MOD13Q1.061": {
            "ndvi": [mask_qa]
            },
        "MYD13Q1.061": {
            "ndvi": [mask_qa]
            },
        "MOD11A1.061": {
            "lst": [mask_bitwise_qa, expand_time_dim]
            },
        "MYD11A1.061": {
            "lst": [mask_bitwise_qa, expand_time_dim]
            },
        "MCD43A3.061": {
            "r0": [
                    shortwave_r0, 
                    partial(mask_qa, masker = ("r0_qa", 1.)),
                    ]
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                variables = None, post_processors = None):
    """Download MODIS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """

    folder = os.path.join(folder, "MODIS")

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = {"MYD13Q1.061": False, "MYD11A1.061": True, "MOD11A1.061": True, "MCD43A3.061": True, "MOD13Q1.061": False}[product_name]
    if spatial_buffer:
        latlim = [latlim[0] - 0.01, latlim[1] + 0.01]
        lonlim = [lonlim[0] - 0.01, lonlim[1] + 0.01]

    if product_name == "MOD13Q1.061" or product_name == "MYD13Q1.061":
        timedelta = np.timedelta64(8, "D")
        timelim[0] = timelim[0] - pd.Timedelta(timedelta)
    elif product_name == "MCD43A3.061":
        timedelta = np.timedelta64(12, "h")
        timelim[0] = timelim[0] - pd.Timedelta(timedelta)
    else:
        timedelta = None

    tiles = tiles_intersect(latlim, lonlim)
    coords = {"x": ["XDim", lonlim], "y": ["YDim", latlim], "t": ["time",timelim]}
    
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = get_crss("MODIS")
    parallel = False
    spatial_tiles = True
    un_pw = accounts.get("NASA")
    request_dims = True
    ds = opendap.download(fn, product_name, coords, 
                variables, post_processors, fn_func, url_func, un_pw = un_pw, 
                tiles = tiles, data_source_crs = data_source_crs, parallel = parallel, 
                spatial_tiles = spatial_tiles, request_dims = request_dims,
                timedelta = timedelta)

    return ds[req_vars_orig]

if __name__ == "__main__":

    products = [
        # ('MCD43A3.061', ["r0"]),
        ('MOD11A1.061', ["lst"]),
        # ('MYD11A1.061', ["lst"]),
        # ('MOD13Q1.061', ["ndvi"]),
        # ('MYD13Q1.061', ["ndvi"]),
    ]

    folder = r"/Users/hmcoerver/Downloads/pywapor_test"
    # latlim = [26.9, 33.7]
    # lonlim = [25.2, 37.2]
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = [datetime.date(2020, 7, 1), datetime.date(2020, 7, 11)]


    for product_name, req_vars in products:
        ds = download(folder, latlim, lonlim, timelim, product_name, req_vars)
        print(ds.rio.crs, ds.rio.grid_mapping)


import os
import json
import pywapor
import getpass
import sys
import requests
import cdsapi
from pywapor.general.logger import log, adjust_logger
from cryptography.fernet import Fernet
from pywapor.collect.product.LANDSAT import espa_api

PASSWORD_INSTRUCTIONS = {
"NASA": """> Used for `MODIS`, `SRTM` and `MERRA2` data.
> Create an account at https://urs.earthdata.nasa.gov.
> Make sure to accept the terms of use at "Applications > Authorized Apps > Approve More Applications":
  * NASA GESDISC DATA ARCHIVE
  * LP DAAC OPeNDAP""",

"TERRA": """> Used for `TERRA` (VITO:PROVA-V).
> Create an account at https://viewer.terrascope.be.""",

"ECMWF": """> Used for `ERA5` (old, please use `"CDS"` instead).
> Create an account at https://cds.climate.copernicus.eu.
  * On your profile page, scroll to the "API key" section.
  * Accept conditions when running `setup("ECMWF")` for the first time.""",

"CDS": """> Used for `ERA5`.
> Create an account at https://cds-beta.climate.copernicus.eu.
  * On your profile page, scroll to the "Personal Access Token" section.
  * Accept conditions when running `setup("CDS")` for the first time.""",

"EARTHEXPLORER": """> Used for `LANDSAT`.
> Create an account at https://earthexplorer.usgs.gov.""",

"COPERNICUS_DATA_SPACE": """> Used for `SENTINEL2` and `SENTINEL3`.
> Create an account at https://dataspace.copernicus.eu.""",

"VIIRSL1": """> Used for `VIIRSL1`
> Normally you do not need this account: it is
> only used when using the `requests.get` download method, 
> which is not the default!
> Create an account at https://ladsweb.modaps.eosdis.nasa.gov/.
> In the top right, press "Login > Generate Token".""",

"LSASAF": """> Used for `MSG_MDSSFTD` and `MSG_MDIDSSF`.
> Create an account at https://mokey.lsasvcs.ipma.pt/.""",
}

def ask_pw(account):
    log.info(f"--> Setting up `{account}` account.").add()
    instructions = PASSWORD_INSTRUCTIONS.get(account, "")
    for line in instructions.split("\n"):
        log.info(line)
    if account == "WAPOR" or account == "VIIRSL1":
        account_name = ""
        pwd = input(f"{account} API token: ")
    elif account == "ECMWF":
        account_name = 'https://cds.climate.copernicus.eu/api/v2'
        api_key_1 = input(f"{account} UID: ")
        api_key_2 = input(f"{account} CDS API key: ")
        pwd = f"{api_key_1}:{api_key_2}"
    elif account == "CDS":
        account_name = 'https://cds-beta.climate.copernicus.eu/api'
        pat = input(f"{account} Personal Access Token: ")
        pwd = f"{pat}"
    else:
        account_name = input(f"{account} username: ")
        pwd = getpass.getpass(f"{account} password: ")  
    log.sub()          
    return account_name, pwd

def setup(account):
    """Asks, saves and tests a username/password combination for `account`.

    Parameters
    ----------
    account : {"NASA" | "TERRA" | "ECMWF" | "CDS" | "COPERNICUS_DATA_SPACE" | "EARTHEXPLORER" | "VIIRSL1" | "LSASAF"}
        Which un/pw combination to store.
    """

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    filehandle = os.path.join(folder, filename)
    json_filehandle = os.path.join(folder, "keys.json")

    if not os.path.exists(filehandle):
        create_key()
    
    with open(filehandle,"r") as f:
        key = f.read()

    cipher_suite = Fernet(key.encode('utf-8'))

    if os.path.exists(json_filehandle):
        # open json file  
        with open(json_filehandle) as f:
            datastore = f.read()
        obj = json.loads(datastore)      
    else:
        obj = {}

    n = 1
    max_attempts = 3
    succes = False

    while n <= max_attempts and not succes:

        account_name, pwd = ask_pw(account)

        log.info(f"--> Testing {account} un/pw.")
        succes, error = {
            "NASA": nasa_account,
            "EARTHEXPLORER": earthexplorer_account,
            "TERRA": terra_account,
            "ECMWF": ecmwf_account,
            "CDS": cds_account,
            "COPERNICUS_DATA_SPACE": copernicus_data_space_account,
            "VIIRSL1": viirs_account,
            "LSASAF": lsasaf_account,
        }[account]((account_name, pwd))

        if succes:
            log.info(f"--> {account} un/pw working.")
            # Encrypting un/pw
            account_name_crypt = cipher_suite.encrypt(("%s" %account_name).encode('utf-8'))
            pwd_crypt = cipher_suite.encrypt(("%s" %pwd).encode('utf-8'))
            obj[account] = ([str(account_name_crypt.decode("utf-8")), str(pwd_crypt.decode("utf-8"))])
            # Save to disk
            with open(json_filehandle, 'w') as outfile:
                json.dump(obj, outfile)     
        else:
            log.warning(f"--> ({n}/{max_attempts}) {account} not working ({error}).")

        n += 1

    if not succes:
        sys.exit(f"Please fix your {account} account.") 

    return

def get(account):
    """Loads a required username/password.

    Parameters
    ----------
    account : {"NASA" | "TERRA" | "ECMWF" | "CDS" | "COPERNICUS_DATA_SPACE" | "EARTHEXPLORER" | "VIIRSL1 | "LSASAF"}
        Which un/pw combination to load.
    """

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    key_file = os.path.join(folder, filename)
    json_file = os.path.join(folder, "keys.json")
    
    if not os.path.exists(key_file):
        create_key()
        if os.path.exists(json_file):
            print("removing old/invalid json file")
            try:
                os.remove(json_file)
            except PermissionError:
                log.warning("--> Unable to remove existing `keys.json` file (PermissionError), please remove the file manually before proceeding.")
    
    if not os.path.exists(json_file):
        setup(account)

    f = open(key_file,"r")
    key = f.read()
    f.close()
    
    cipher_suite = Fernet(key.encode('utf-8'))    
    
    # open json file  
    with open(json_file) as f:
        datastore = f.read()
    obj = json.loads(datastore)      
    f.close()

    if account not in obj.keys():
        setup(account)
        obj = None 
        with open(json_file) as f:
            datastore = f.read()
        obj = json.loads(datastore)      
        f.close()       

    username_crypt, pwd_crypt = obj[account]
    username = cipher_suite.decrypt(username_crypt.encode("utf-8"))
    pwd = cipher_suite.decrypt(pwd_crypt.encode("utf-8"))  

    return (str(username.decode("utf-8")), str(pwd.decode("utf-8")))

def create_key():
    """Generates a key file.
    """
    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    filename = "secret.txt"
    filehandle = os.path.join(folder, filename)

    if os.path.exists(filehandle):
        try:
            os.remove(filehandle)
        except PermissionError:
            log.warning("--> Unable to remove existing `secret.txt` file (PermissionError), please remove the file manually before proceeding.")
        
    with open(filehandle,"w+") as f:
        f.write(str(Fernet.generate_key().decode("utf-8")))

def lsasaf_account(user_pw):

    test_url = r"https://thredds.lsasvcs.ipma.pt/thredds/catalog/catalog.html"

    session = requests.Session()
    session.auth = user_pw
    x = session.get(test_url)

    if not x.ok:
        error = str(x.status_code)
        succes = False
    else:
        error = ""
        succes = True

    return succes, error


def terra_account(user_pw):
    """Check if the given or stored TERA username/password combination
    is correct. Accounts can be created on https://viewer.terrascope.be.

    Parameters
    ----------
    user_pw : tuple, optional
        ("username", "password") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """

    url = "https://sso.vgt.vito.be/auth/realms/terrascope/protocol/openid-connect/token"
    params = {
        "grant_type": "password",
        "client_id": "public",
        "username": user_pw[0],
        "password": user_pw[1],
    }
    headers = {'content-type': 'application/x-www-form-urlencoded'}

    x = requests.post(url, data = params, headers = headers)

    error = ""

    if not x.ok:
        error = eval(x.text)["error_description"]
        succes = False
    else:
        succes = True

    return succes, error

def nasa_account(user_pw):
    """Check if the given or stored NASA username/password combination is 
    correct. Accounts can be created on https://urs.earthdata.nasa.gov/users/new.

    Parameters
    ----------
    user_pw : tuple, optional
        ("username", "password") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """

    folder = os.path.dirname(os.path.realpath(pywapor.__path__[0]))
    test_url = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2T3NXGLC.5.12.4/1987/08/MERRA2_100.tavg3_2d_glc_Nx.19870801.nc4"
    test_file = os.path.join(folder, "nasa_test.nc4")

    if os.path.isfile(test_file):
        try:
            os.remove(test_file)
        except PermissionError:
            ...

    username, password = user_pw

    try:
        x = requests.get(test_url, allow_redirects = False)
    except requests.exceptions.ConnectionError as e:
        # TODO fix this...
        log.add().warning("> Test not working right now, assuming your account is valid.")
        log.warning("> If you want to enter a new password, run `pywapor.collect.accounts.setup('NASA')`.").sub()
        succes = True
        return succes, ""

    y = requests.get(x.headers['location'], auth = (username, password))

    error = ""

    if x.ok and y.ok:

        with open(test_file, 'w+b') as z:
            z.write(y.content)
    
        if os.path.isfile(test_file):
            statinfo = os.stat(test_file)
            succes = statinfo.st_size == 3963517
            if not succes:
                error = "please add 'NASA GESDISC DATA ARCHIVE' to 'Approved Applications'."
        else:
            succes = False

    else:
        error = "wrong username/password."
        succes = False
    
    if os.path.isfile(test_file):
        try:
            os.remove(test_file)
        except PermissionError:
            ...

    return succes, error

def viirs_account(user_pw):
    """Check if the given or stored VIIRSL1 token is 
    correct. Accounts can be created on https://ladsweb.modaps.eosdis.nasa.gov/.
    Then in the top right, press "Login" and "generate token".

    Parameters
    ----------
    user_pw : tuple, optional
        ("", "token") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """

    _, token = user_pw

    headers = {'Authorization': 'Bearer ' + token}
    test_url = "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/5110/CLDMSK_L2_VIIRS_SNPP/2023/060/CLDMSK_L2_VIIRS_SNPP.A2023060.0006.001.2023233192108.nc"
    response = requests.Session().get(test_url, headers=headers, stream = True)
    for data in response.iter_content(chunk_size=1024):
        succes = b'DOCTYPE' not in data
        if not succes:
            error = "wrong token."
        else:
            error = ""
        break

    return succes, error


def earthexplorer_account(user_pw):
    """Check if the given or stored WAPOR token is 
    correct. Accounts can be created on https://wapor.apps.fao.org/home/WAPOR_2/1.

    Parameters
    ----------
    user_pw : tuple, optional
        ("", "token") to check, if `None` will try to load the 
        password from the keychain, by default None.

    Returns
    -------
    bool
        True if the password works, otherwise False.
    """

    username, pw = user_pw

    try:
        response = espa_api('user', uauth = (username, pw))
    except requests.exceptions.ConnectionError as e:
        # TODO fix this...
        log.add().warning("> Test not working right now, assuming your account is valid.")
        log.warning("> If you want to enter a new password, run `pywapor.collect.accounts.setup('EARTHEXPLORER')`.").sub()
        succes = True
        return succes, ""

    if isinstance(response, type(None)):
        response = {"username": "x"}

    if "email" in response.keys() and response["username"] == username:
        succes = True
        error = ""
    elif response["username"].casefold() == username.casefold() and response["username"] != username:
        error = "wrong username, please make sure capitalization is correct."
        succes = False
    else:
        error = "wrong username/password."
        succes = False

    return succes, error

def copernicus_data_space_account(user_pw):
    """Check if the given or stored COPERNICUS_DATA_SPACE account is
    correct. Accounts can be created on https://dataspace.copernicus.eu

    Parameters
    ----------
    user_pw : tuple
        ("username", "password") to check.

    Returns
    -------
    tuple
        First item is `True` if the password works, otherwise `False`,
        seconds item is a error message.
    """

    username, password = user_pw

    try:

        data = {
            "client_id": "cdse-public",
            "username": username,
            "password": password,
            "grant_type": "password",
            }
        
        r = requests.post("https://identity.dataspace.copernicus.eu/auth/realms/CDSE/protocol/openid-connect/token",
                            data=data,
                            )
        r.raise_for_status()
        token = r.json()

        if "access_token" in token.keys():
            succes = True
            error = ""
        else:
            succes = False
            error = "Invalid username/password"

    except Exception as e:
        exception_args = getattr(e, "args", ["Invalid username/password"])
        error = exception_args[0]
        succes = False

    return succes, error

def cds_account(user_pw):
    """Check if the given or stored ECMWF key is 
    correct. Accounts can be created on https://cds.climate.copernicus.eu/#!/home.

    Parameters
    ----------
    user_pw : tuple
        ("username", "password") to check.

    Returns
    -------
    tuple
        First item is `True` if the password works, otherwise `False`,
        seconds item is a error message.
    """

    url, key = user_pw

    try:
        vrfy = {"NO": False, "YES": True}.get(os.environ.get("PYWAPOR_VERIFY_SSL", "YES"), True)
        c = cdsapi.Client(url = url, key = key, verify = vrfy, quiet = True)
        fp = os.path.join(pywapor.__path__[0], "test.zip")
        _ = c.retrieve(
            'sis-agrometeorological-indicators',
            {
                'format': 'zip',
                'variable': '2m_temperature',
                'statistic': '24_hour_mean',
                'year': '2005',
                'month': '11',
                'day': ['01'],
                'area': [2,-2,-2,2],
                'version': '1_1'
            },
            fp)
        try:
            os.remove(fp)
        except PermissionError:
            ... # windows...
        succes = True
        error = ""
    except Exception as e:
        exception_args = getattr(e, "args", ["wrong_token"])
        succes = False
        if len(exception_args) > 0:
            if "Client has not agreed to the required terms and conditions" in str(exception_args[0]):
                error = exception_args[0]
            else:
                error = str(e)
        else:
            error = "wrong key"

    return succes, error

def ecmwf_account(user_pw):
    """Check if the given or stored ECMWF key is 
    correct. Accounts can be created on https://cds.climate.copernicus.eu/#!/home.

    Parameters
    ----------
    user_pw : tuple
        ("username", "password") to check.

    Returns
    -------
    tuple
        First item is `True` if the password works, otherwise `False`,
        seconds item is a error message.
    """

    url, key = user_pw

    try:
        vrfy = {"NO": False, "YES": True}.get(os.environ.get("PYWAPOR_VERIFY_SSL", "YES"), True)
        c = cdsapi.Client(url = url, key = key, verify = vrfy, quiet = True)
        fp = os.path.join(pywapor.__path__[0], "test.zip")
        _ = c.retrieve(
            'sis-agrometeorological-indicators',
            {
                'format': 'zip',
                'variable': '2m_temperature',
                'statistic': '24_hour_mean',
                'year': '2005',
                'month': '11',
                'day': ['01'],
                'area': [2,-2,-2,2],
                'version': '1_1'
            },
            fp)
        try:
            os.remove(fp)
        except PermissionError:
            ... # windows...
        succes = True
        error = ""
    except Exception as e:
        exception_args = getattr(e, "args", ["wrong_token"])
        succes = False
        if len(exception_args) > 0:
            if "Client has not agreed to the required terms and conditions" in str(exception_args[0]):
                error = exception_args[0]
        else:
            error = "wrong key"

    return succes, error

if __name__ == "__main__":
    ...
    adjust_logger(True, r"/Users/hmcoerver/Desktop", "INFO")

    un_pw1= get("NASA")
    # print("NASA", un_pw1)
    # check1 = nasa_account(un_pw1)

    un_pw2 = get("TERRA")
    # print("TERRA", un_pw2)
    # check2 = terra_account(un_pw2)

    un_pw4 = get("ECMWF")
    # print("ECMWF", un_pw4)
    # check4 = ecmwf_account(un_pw4)

    un_pw6 = get("EARTHEXPLORER")
    # print("EARTHEXPLORER", un_pw6)
    # check6 = earthexplorer_account(un_pw6)

    un_pw7 = get("COPERNICUS_DATA_SPACE")
    # print("COPERNICUS_DATA_SPACE", un_pw7)
    # check7 = copernicus_data_space_account(un_pw7)

    un_pw8 = get("VIIRSL1")
    # print("VIIRSL1", un_pw8)
    # check8 = viirs_account(un_pw8)

    un_pw8 = get("LSASAF")
    # print("VIIRSL1", un_pw8)
    # check8 = viirs_account(un_pw8)

    un_pw9 = get("CDS")

    # print(check1, check2, check4, check6, check7, check8)


# -*- coding: utf-8 -*-
"""
Code to run the SERoot model.
"""
import xarray as xr
import os
import numpy as np
import warnings
import pandas as pd
import datetime
import pywapor.general as g
import pywapor.et_look_v2_v3 as ETLook_v2_v3
from pywapor.general.logger import log, adjust_logger
from pywapor.general.processing_functions import save_ds, open_ds
import copy
import pywapor.pre_se_root as pre_se_root
from pywapor.general import levels

def se_root(folder, latlim, lonlim, timelim, sources = "level_1", bin_length = "DEKAD", se_root_version = None, **kwargs):
    """Runs `pre_se_root` and `se_root`, used internally to generate the `se_root` data in `pre_et_look`.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    sources : str | dict
        Configuration for each variable and source.
    bin_length : int | "DEKAD"
        Composite length.

    Returns
    -------
    xr.Dataset
        Dataset with `se_root` variable.
    """
    if isinstance(se_root_version, type(None)) and isinstance(sources, str):
        se_root_version = {True: "v2", False: "v3"}["v3" not in sources]
    elif isinstance(se_root_version, type(None)) and isinstance(sources, dict):
        se_root_version = "v2"
        log.warning("--> Using se_root v2.")

    if isinstance(sources, str):
        sources = levels.pre_se_root_levels(sources)

    ds_in = pre_se_root.main(folder, latlim, lonlim, timelim, sources, bin_length, buffer_timelim = False)
    ds_out = main(ds_in, se_root_version = se_root_version)

    return ds_out

def main(input_data, se_root_version = "v2", export_vars = "default", chunks = {"time": -1, "x": 500, "y": 500}):
    """Run the `se_root` model.

    Parameters
    ----------
    input_data : str | xr.Dataset
        (Path to) dataset generated by `pywapor.pre_se_root`.
    se_root_version : "v2" | "v3", optional
        Which version of the SERoot model to use, by default "v2".
    export_vars : "default" | "all" | list, optional
        Specify which variables to save inside the output file. `"Default"` only 
        stores `se_root`. `"all"` stores all calculated variables. Use a
        list to specify a custom output set, by default "default".
    chunks : dict, optional
        Specify how the calculations are split up. Increase chunk sizes to speed up calculation, 
        decrease to use less RAM, by default {"time": 1, "x": 1000, "y": 1000}.

    Returns
    -------
    xr.Dataset
        Outputs from the model.
    """

    # Inputs
    if isinstance(input_data, str):
        ds = open_ds(input_data, chunks = chunks)
    else:
        ds = input_data.chunk(chunks)
        input_data = ds.encoding["source"]

    _ = adjust_logger(True, os.path.split(input_data)[0], "INFO")

    t1 = datetime.datetime.now()
    log.info("> SE_ROOT").add()

    # Version
    ETLook = ETLook_v2_v3

    log.info(f"--> Running `se_root` ({se_root_version}).").add()

    # Allow skipping of et_look-functions if not all of its required inputs are
    # available.
    g.lazifier.decorate_submods(ETLook, g.lazifier.etlook_decorator)

    ds = g.variables.initiate_ds(ds)

    fp, fn = os.path.split(input_data)

    if not ds["v2m_i"].dtype == object and not ds["u2m_i"].dtype == object:
        ds["u_i"] = np.sqrt(ds["v2m_i"]**2 + ds["u2m_i"]**2)

    doy = [int(pd.Timestamp(x).strftime("%j")) for x in ds["time"].values]
    ds["doy"] = xr.DataArray(doy, coords = ds["time"].coords).chunk("auto")
    dtime = [pd.Timestamp(x).hour + (pd.Timestamp(x).minute / 60) for x in ds["time"].values]
    ds["dtime"] = xr.DataArray(dtime, coords = ds["time"].coords).chunk("auto")
    ds["sc"] = ETLook.solar_radiation.seasonal_correction(ds["doy"])
    ds["decl"] = ETLook.solar_radiation.declination(ds["doy"])
    ds["day_angle"] = ETLook.clear_sky_radiation.day_angle(ds["doy"])

    ds["p_air_i"] = ETLook.meteo.air_pressure_kpa2mbar(ds["p_air_i"])
    ds["p_air_0_i"] = ETLook.meteo.air_pressure_kpa2mbar(ds["p_air_0_i"])

    ds["vc"] = ETLook.leaf.vegetation_cover(ds["ndvi"], nd_min = ds["nd_min"], nd_max = ds["nd_max"], vc_pow = ds["vc_pow"])

    ds["t_air_k_i"] = ETLook.meteo.air_temperature_kelvin_inst(ds["t_air_i"])

    ds["vp_i"] = ETLook.meteo.vapour_pressure_from_specific_humidity_inst(ds["qv_i"], ds["p_air_i"])
    if ds["vp_i"].dtype == object:
        ds["vp_i"] = ETLook.meteo.vapour_pressure_from_dewpoint_inst(ds["t_dew_i"])

    ds["qv_i"] = ETLook.meteo.specific_humidity_from_vapour_pressure(ds["vp_i"], ds["p_air_i"])
    
    ds["ad_moist_i"] = ETLook.meteo.moist_air_density_inst(ds["vp_i"], ds["t_air_k_i"])
    ds["ad_dry_i"] = ETLook.meteo.dry_air_density_inst(ds["p_air_i"], ds["vp_i"], ds["t_air_k_i"])
    ds["ad_i"] = ETLook.meteo.air_density_inst(ds["ad_dry_i"], ds["ad_moist_i"])
    ds["u_b_i_bare"] = ETLook.soil_moisture.wind_speed_blending_height_bare(ds["u_i"], z0m_bare = ds["z0m_bare"], z_obs = ds["z_obs"], z_b = ds["z_b"])
    ds["lon_rad"] = ETLook.solar_radiation.longitude_rad(ds["x"]).chunk("auto")
    ds["lat_rad"] = ETLook.solar_radiation.latitude_rad(ds["y"]).chunk("auto")
    ds["ha"] = ETLook.solar_radiation.hour_angle(ds["sc"], ds["dtime"], ds["lon_rad"])

    ds["ied"] = ETLook.clear_sky_radiation.inverse_earth_sun_distance(ds["day_angle"])
    ds["h0"] = ETLook.clear_sky_radiation.solar_elevation_angle(ds["lat_rad"], ds["decl"], ds["ha"])
    ds["h0ref"] = ETLook.clear_sky_radiation.solar_elevation_angle_refracted(ds["h0"])
    ds["m"] = ETLook.clear_sky_radiation.relative_optical_airmass(ds["p_air_i"], ds["p_air_0_i"], ds["h0ref"])
    ds["rotm"] = ETLook.clear_sky_radiation.rayleigh_optical_thickness(ds["m"])
    ds["Tl2"] = ETLook.clear_sky_radiation.linke_turbidity(ds["wv_i"], ds["aod550_i"], ds["p_air_i"], ds["p_air_0_i"])
    ds["G0"] = ETLook.clear_sky_radiation.extraterrestrial_irradiance_normal(ds["IO"], ds["ied"])
    ds["B0c"] = ETLook.clear_sky_radiation.beam_irradiance_normal_clear(ds["G0"], ds["Tl2"], ds["m"], ds["rotm"], ds["h0"])
    ds["Bhc"] = ETLook.clear_sky_radiation.beam_irradiance_horizontal_clear(ds["B0c"], ds["h0"])
    ds["Dhc"] = ETLook.clear_sky_radiation.diffuse_irradiance_horizontal_clear(ds["G0"], ds["Tl2"], ds["h0"])

    ds["ra_hor_clear_i"] = ETLook.clear_sky_radiation.ra_clear_horizontal(ds["Bhc"], ds["Dhc"])
    ds["emiss_atm_i"] = ETLook.soil_moisture.atmospheric_emissivity_inst(ds["vp_i"], ds["t_air_k_i"])

    ds["rn_bare"] = ETLook.soil_moisture.net_radiation_bare(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["lst"], ds["r0_bare"])
    ds["rn_full"] = ETLook.soil_moisture.net_radiation_full(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["lst"], ds["r0_full"])

    ds["h_bare"] = ETLook.soil_moisture.sensible_heat_flux_bare(ds["rn_bare"], fraction_h_bare = ds["fraction_h_bare"])
    ds["h_full"] = ETLook.soil_moisture.sensible_heat_flux_full(ds["rn_full"], fraction_h_full = ds["fraction_h_full"])
    ds["u_b_i_full"] = ETLook.soil_moisture.wind_speed_blending_height_full_inst(ds["u_i"], z0m_full = ds["z0m_full"], z_obs = ds["z_obs"], z_b = ds["z_b"])

    ds["u_star_i_bare"] = ETLook.soil_moisture.friction_velocity_bare_inst(ds["u_b_i_bare"], z0m_bare = ds["z0m_bare"], disp_bare = ds["disp_bare"], z_b = ds["z_b"])
    ds["u_star_i_full"] = ETLook.soil_moisture.friction_velocity_full_inst(ds["u_b_i_full"], z0m_full = ds["z0m_full"], disp_full = ds["disp_full"], z_b = ds["z_b"])
    ds["L_bare"] = ETLook.soil_moisture.monin_obukhov_length_bare(ds["h_bare"], ds["ad_i"], ds["u_star_i_bare"], ds["t_air_k_i"])
    ds["L_full"] = ETLook.soil_moisture.monin_obukhov_length_full(ds["h_full"], ds["ad_i"], ds["u_star_i_full"], ds["t_air_k_i"])

    ds["u_i_soil"] = ETLook.soil_moisture.wind_speed_soil_inst(ds["u_i"], ds["L_bare"], z_obs = ds["z_obs"])
    
    ds["ras"] = ETLook.soil_moisture.aerodynamical_resistance_forced_convection_soil(ds["u_i_soil"])
    ds["raa"] = ETLook.soil_moisture.aerodynamical_resistance_forced_convection_bare(ds["u_i"], ds["L_bare"], z0m_bare = ds["z0m_bare"], disp_bare = ds["disp_bare"], z_obs = ds["z_obs"])
    ds["rac"] = ETLook.soil_moisture.aerodynamical_resistance_forced_convection_full(ds["u_i"], ds["L_full"], z0m_full = ds["z0m_full"], disp_full = ds["disp_full"], z_obs = ds["z_obs"])
    
    if se_root_version == "v3":

        ds["rah_bare_free"] = ETLook.soil_moisture.aerodynamical_resistance_free_convection_bare(ds["h_bare"], ds["t_air_k_i"], ds["ad_i"], z0m_bare = ds["z0m_bare"])
        ds["rah_full_free"] = ETLook.soil_moisture.aerodynamical_resistance_free_convection_full(ds["h_full"], ds["t_air_k_i"], ds["ad_i"], z0m_full = ds["z0m_full"])
        
        ds["raa"] = ETLook.soil_moisture.aerodynamical_resistance_bare(ds["raa"], ds["ras"], rah_bare_free = ds["rah_bare_free"])
        ds["rac"] = ETLook.soil_moisture.aerodynamical_resistance_full(ds["rac"], rah_full_free = ds["rah_full_free"])

    ds["t_max_bare"] = ETLook.soil_moisture.maximum_temperature_bare(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["ad_i"], ds["raa"], ds["ras"], ds["r0_bare"])
    ds["t_max_full"] = ETLook.soil_moisture.maximum_temperature_full(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["ad_i"], ds["rac"], ds["r0_full"])

    ds["t_wet_i"] = ETLook.soil_moisture.wet_bulb_temperature_inst_new(ds["t_air_i"], ds["qv_i"], ds["p_air_i"])
    ds["lst_max"] = ETLook.soil_moisture.maximum_temperature(ds["t_max_bare"], ds["t_max_full"], ds["vc"])

    ds["t_wet_k_i"] = ETLook.meteo.wet_bulb_temperature_kelvin_inst(ds["t_wet_i"])
    ds["lst_min"] = ETLook.soil_moisture.minimum_temperature(ds["t_wet_k_i"], ds["t_air_k_i"], ds["vc"])

    ds["se_root"] = ETLook.soil_moisture.soil_moisture_from_maximum_temperature(ds["lst_max"], ds["lst"], ds["lst_min"])

    if export_vars == "all":
        ...
    elif export_vars == "default":
        keep_vars = [
            'se_root'
            ]
        keep_vars = [x for x in keep_vars if x in ds.variables]
        ds = ds[keep_vars]
    elif isinstance(export_vars, list):
        keep_vars = copy.copy(export_vars)
        ds = ds[keep_vars]
    else:
        raise ValueError(f"Please provide a valid `export_vars` ('all', 'default' or a list).")

    ds = g.variables.fill_attrs(ds)

    log.sub()
    if len(ds.data_vars) == 0:
        log.info("--> No data to export, try adjusting `export_vars`.")
        ds = None
    else:
        fn = fn.replace("in", "out")
        fp_out = os.path.join(fp, fn)
        while os.path.isfile(fp_out):
            fp_out = fp_out.replace(".nc", "_.nc")

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="invalid value encountered in power")
            warnings.filterwarnings("ignore", message="invalid value encountered in log")
            ds = save_ds(ds, fp_out, encoding = "initiate", chunks = chunks, label = f"Saving output to `{os.path.split(fp_out)[-1]}`.")

    t2 = datetime.datetime.now()
    log.sub().info(f"< SE_ROOT ({str(t2 - t1)})")

    return ds

def test_ds(ds, var):
    from dask.diagnostics import Profiler, ResourceProfiler, CacheProfiler, ProgressBar
    from dask.diagnostics import visualize

    with Profiler() as prof, ResourceProfiler(dt=0.25) as rprof, CacheProfiler() as cprof, ProgressBar():
        out = ds[var].compute()

    visualize([prof, rprof, cprof])
    return out

if __name__ == "__main__":

    se_root_version = "v2"
    export_vars = "default"
    chunks = {"time": 1, "x": 1000, "y": 1000}
    # input_data = r"/Users/hmcoerver/Local/20220325_20220415_test_data/se_root_in.nc"

    # out = main(input_data, se_root_version = se_root_version, export_vars = export_vars)



from pywapor.collect.protocol import opendap
from pywapor.collect.protocol.projections import get_crss
import os
import xarray as xr
from pywapor.general.processing_functions import open_ds
from pywapor.enhancers.temperature import kelvin_to_celsius
from pywapor.enhancers.wind import adjust_wind_height
from functools import partial
import numpy as np
from pywapor.enhancers.pressure import pa_to_kpa
import copy

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "inst3_2d_asm_Nx": {
                    "t2m": [("time", "lat", "lon"), "t_air"],
                    "u2m": [("time", "lat", "lon"), "u2m"],
                    "v2m": [("time", "lat", "lon"), "v2m"],
                    "qv2m": [("time", "lat", "lon"), "qv"],
                    "tqv": [("time", "lat", "lon"), "wv"],
                    "ps": [("time", "lat", "lon"), "p_air"],
                    "slp": [("time", "lat", "lon"), "p_air_0"],
                        },

        "tavg1_2d_slv_Nx": {
                    "t2m": [("time", "lat", "lon"), "t_air"],
                    "qv2m": [("time", "lat", "lon"), "qv"],
                    "u10m": [("time", "lat", "lon"), "u10m"],
                    "v10m": [("time", "lat", "lon"), "v10m"],
                    "ps": [("time", "lat", "lon"), "p_air"],
                    "slp": [("time", "lat", "lon"), "p_air_0"],
                    "tqv": [("time", "lat", "lon"), "wv"],
                    "to3": [("time", "lat", "lon"), "to3"],
                        },

        "tavg1_2d_rad_Nx": {
                    "swgdn": [("time", "lat", "lon"), "ra_flat"], # W/m2, surface incoming shortwave flux
                        },

        "tavg1_2d_lnd_Nx": {
            "prectot": [("time", "lat", "lon"), "p"],
        },

        "tavg3_2d_aer_Nx": {
            "totangstr": [("time", "lat", "lon"), "totangstr"],
            "totexttau": [("time", "lat", "lon"), "totexttau"]
        }

    }

    req_dl_vars = {
        "inst3_2d_asm_Nx": {
            "t_air": ["t2m"],
            "t_air_max": ["t2m"],
            "t_air_min": ["t2m"],
            "u2m": ["u2m"],
            "v2m": ["v2m"],
            "qv": ["qv2m"],
            "wv": ["tqv"],
            "p_air": ["ps"],
            "p_air_0": ["slp"],
        },
        "tavg1_2d_slv_Nx": {
            "t_air": ["t2m"],
            "t_air_max": ["t2m"],
            "t_air_min": ["t2m"],
            "u10m": ["u10m"],
            "v10m": ["v10m"],
            "u2m": ["u10m"],
            "v2m": ["v10m"],
            "qv": ["qv2m"],
            "wv": ["tqv"],
            "p_air": ["ps"],
            "p_air_0": ["slp"],
            "to3": ["to3"], # total_column_ozone
        },
        "tavg1_2d_rad_Nx": {
            "ra_flat": ["swgdn"], # surface_incoming_shortwave_flux
        },
        "tavg1_2d_lnd_Nx": {
            "p": ["prectot"],
        },
        "tavg3_2d_aer_Nx": {
            "totangstr": ["totangstr"], # total aerosol angstrom parameter [470-870 nm] 
            "totexttau": ["totexttau"], # total aerosol extinction aot [550 nm] 
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "inst3_2d_asm_Nx": {
            "t_air": [kelvin_to_celsius], 
            "t_air_max": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_max")],
            "t_air_min": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_min")],
            "u2m": [],
            "v2m": [],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
        },
        "tavg1_2d_slv_Nx": {
            "t_air": [kelvin_to_celsius],
            "t_air_max": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_max")],
            "t_air_min": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_min")],
            "v10m": [],
            "u10m": [],
            "u2m": [adjust_wind_height],
            "v2m": [adjust_wind_height],
            "qv": [],
            "wv": [],
            "p_air": [pa_to_kpa],
            "p_air_0": [pa_to_kpa],
            "to3": [],
        },
        "tavg1_2d_rad_Nx": {
            "ra_flat": [],
        },
        "tavg1_2d_lnd_Nx": {
            "p": [],
        },
        "tavg3_2d_aer_Nx": {
            "totangstr": [],
            "totexttau": [],
        },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                 variables = None, post_processors = None):
    """Download GEOS5 data and store it in a single netCDF file. Product
    docs are here https://gmao.gsfc.nasa.gov/pubs/docs/Lucchesi1202.pdf

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "GEOS5")

    if not os.path.exists(folder):
        os.makedirs(folder)

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.25, latlim[1] + 0.25]
        lonlim = [lonlim[0] - 0.3125, lonlim[1] + 0.3125]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    data_source_crs = get_crss("WGS84")

    url = f"https://opendap.nccs.nasa.gov/dods/GEOS-5/fp/0.25_deg/assim/{product_name}"

    timedelta = np.timedelta64(90, "m")

    ds = opendap.download_xarray(url, fn, coords, variables, post_processors, 
                                    data_source_crs = data_source_crs,
                                    timedelta = timedelta)

    return ds[req_vars_orig]

if __name__ == "__main__":

    from pywapor.general.logger import adjust_logger
    import datetime

    args = {'folder': '/Users/hmcoerver/Local/test_dl_GEOS5_0',
            'latlim': [29.4, 29.7],
            'lonlim': [30.7, 31.0],
            'timelim': [datetime.date(2022, 3, 1), datetime.date(2022, 3, 3)],
            'product_name': 'inst3_2d_asm_Nx',
            'req_vars': ['t_air',
            't_air_max',
            't_air_min',
            'u2m',
            'v2m',
            'qv',
            'wv',
            'p_air',
            'p_air_0']}
    
    variables = None
    post_processors = None

    for var, value in args.items():
        if isinstance(value, str):
            exec(f"{var} = '{value}'")
        else:
            exec(f"{var} = {value}")

    # adjust_logger(True, folder, "INFO")




# import numpy as np

# def get_units(dss):
#     """Get an overview of the units of different variables inside xr.Datasets.

#     Parameters
#     ----------
#     dss : list
#         List with xr.Datasets. Each variable in each dataset is checked for a 
#         `"unit"` attribute.

#     Returns
#     -------
#     dict
#         Dictionary with the units per variable.
#     """
#     units = dict()
#     for sub_ds in dss:
#         variables = list(sub_ds.keys())
#         for var in variables:
#             if hasattr(sub_ds[var], "unit"):
#                 unit = sub_ds[var].unit
#             else:
#                 unit = "unknown"
#             if var in units.keys():
#                 units[var] = np.append(units[var], unit)
#             else:
#                 units[var] = np.array([unit])
#     return units

# def check_units(units, strictness = "low"):
#     """Test whether each variable (as key in `units`) has identical units.

#     Parameters
#     ----------
#     units : dict
#         Dictionary with the units per variable, can be generated with `compositer.get_units`.
#     strictness : {"low" | "medium" | "high}, optional
#         low - Units need to be the same, but 'unknown' units are assumed to be correct.
#         med - Units need to be the same and 'unknown' units are assumed to be different.
#         high - All units must be known and identical.

#     Examples
#     --------
#     >>> units = {"test1": np.array(["C", "C","unknown"]),
#     ...          "test2": np.array(["C", "C", "K"]),
#     ...          "test3": np.array(["C", "C", "C"]),
#     ...          "test4": np.array(["unknown", "unknown", "unknown"])}
#     >>> check_units(units)
#     {'test1': True, 'test2': False, 'test3': True, 'test4': True}
#     >>> check_units(units, strictness = "med")
#     {'test1': False, 'test2': False, 'test3': True, 'test4': True}
#     >>> check_units(units, strictness = "high")
#     {'test1': False, 'test2': False, 'test3': True, 'test4': False}
#     """
#     results = dict()
#     if strictness == "low":
#         for k, v in units.items():
#             check = np.unique(v[v!="unknown"]).size <= 1
#             results[k] = check
#     if strictness == "med":
#         for k, v in units.items():
#             check = np.unique(v).size == 1
#             results[k] = check
#     if strictness == "high":
#         for k, v in units.items():
#             check = np.unique(v).size == 1 and "unknown" not in v
#             results[k] = check
#     return results


"""
Functions to prepare input for `pywapor.se_root`, more specifically to
interpolate various parameters in time to match with land-surface-temperature
times. 
"""
from pywapor.general.processing_functions import save_ds, open_ds, remove_ds, log_example_ds, func_from_string
from pywapor.general.reproject import align_pixels
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.logger import log
import os
import numpy as np
import xarray as xr
from itertools import chain
import pywapor.general.levels as levels
from pywapor.enhancers.smooth.whittaker import whittaker_smoothing

def main(dss, sources, folder, general_enhancers, example_t_vars = ["lst"]):
    """Aligns the datetimes in de `dss` xr.Datasets with the datetimes of the 
    dataset with variable `example_t_var`.

    Parameters
    ----------
    dss : dict
        Keys are tuples of (`source`, `product_name`), values are xr.Dataset's 
        or paths (str) to netcdf files, which will be aligned along the time dimensions.
    sources : dict
        Configuration for each variable and source.
    folder : str
        Path to folder in which to store (intermediate) data.
    general_enhancers : list
        Functions to apply to the xr.Dataset before creating the final
        output, by default "default".
    example_t_var : list, optional
        Which variables to align the other datasets to in the time dimension, by default ["lst"].

    Returns
    -------
    xr.Dataset
        Dataset in which all variables have been interpolated to the same times.
    """
    chunks = {"time": -1, "y": 500, "x": 500}

    # Open unopened netcdf files.
    dss = {**{k: open_ds(v) for k, v in dss.items() if isinstance(v, str)}, 
            **{k:v for k,v in dss.items() if not isinstance(v, str)}}

    # Determine final output path.
    final_path = os.path.join(folder, "se_root_in.nc")

    # Make list to store intermediate datasets.
    dss2 = dict()

    # Make inventory of all variables.
    variables = [x for x in np.unique(list(chain.from_iterable([ds.data_vars for ds in dss.values()]))).tolist() if x in sources.keys()]
    variables = example_t_vars + [x for x in variables if x not in example_t_vars]

    # Create variable to store times to interpolate to.
    example_time = None

    cleanup = list()

    # Loop over the variables
    for var in variables:

        dst_path = os.path.join(folder, "_instantaneous",f"{var}_i.nc")

        if os.path.isfile(dst_path):
            ds = open_ds(dst_path, chunks = chunks)
            if var in example_t_vars:
                if isinstance(example_time, type(None)):
                    example_time = ds["time"]
        else:
            config = sources[var]
            spatial_interp = config["spatial_interp"]
            temporal_interp = config["temporal_interp"]

            if isinstance(temporal_interp, dict):
                kwargs = temporal_interp.copy()
                temporal_interp = kwargs.pop("method")
            else:
                kwargs = {}

            # Align pixels of different products for a single variable together.
            dss_part = [ds[[var]] for ds in dss.values() if var in ds.data_vars]
            if len(dss_part) > 1:
                log.info(f"--> Spatially aligning {len(dss_part)} `{var}` products together.").add()
            dss1, temp_files1 = align_pixels(dss_part, folder, spatial_interp, fn_append = "_step1")
            if len(dss_part) > 1:
                log.sub()
            cleanup.append(temp_files1)

            # Combine different source_products (along time dimension).
            ds = xr.combine_nested(dss1, concat_dim = "time").chunk(chunks).sortby("time").squeeze()

            if var in example_t_vars:
                if "time" not in ds[var].dims:
                    ds[var] = ds[var].expand_dims("time").transpose("time", "y", "x")
                
                if isinstance(example_time, type(None)):
                    example_time = ds["time"]
                else:
                    example_time = xr.concat([example_time, ds["time"]], dim = "time").drop_duplicates("time").sortby("time")
            else:
                ...

            if "time" in ds[var].dims and not isinstance(temporal_interp, type(None)):
                orig_time_size = ds["time"].size

                if temporal_interp == "whittaker":
                    if (not ds.time.equals(example_time)) and (not var in example_t_vars):
                        new_x = example_time.values
                    else:
                        new_x = None
                        
                    if "weights" in kwargs.keys():
                        ds["sensor"] = xr.combine_nested([xr.ones_like(x.time.astype(int)) * i for i, x in enumerate(dss1)], concat_dim="time").sortby("time")
                        source_legend = {str(i): os.path.split(x.encoding["source"])[-1].replace(".nc", "") for i, x in enumerate(dss1)}
                        ds["sensor"] = ds["sensor"].assign_attrs(source_legend)
                    ds = whittaker_smoothing(ds, var, chunks = chunks, new_x = new_x, **kwargs)
                    ds = ds.rename_vars({f"{var}_smoothed": var})
                else:
                    if not ds.time.equals(example_time):
                        new_coords = xr.concat([ds.time, example_time], dim = "time").drop_duplicates("time").sortby("time")
                        ds = ds.reindex_like(new_coords).chunk(chunks)
                    ds = ds.interpolate_na(dim = "time", method = temporal_interp, **kwargs).sel(time = example_time)

                lbl = f"Aligning times in `{var}` ({orig_time_size}) with `{'` and `'.join(example_t_vars)}` ({example_time.time.size}, {temporal_interp})."
                ds = save_ds(ds, dst_path, chunks = chunks, encoding = "initiate", label = lbl)
                log.add().info(f"> shape: {ds[var].shape}, kwargs: {list(kwargs.keys())}.").sub()
            else:
                lbl = f"Saving spatially aligned `{var}` data without temporal interpolation."
                ds = save_ds(ds, dst_path, chunks = chunks, encoding = "initiate", label = lbl)

        # Set dimension order.
        ds = ds.transpose(*sorted(list(ds.sizes.keys()), key = lambda e: ["time", "y", "x"].index(e)))

        dss2[var] = ds

    # Apply variable specific enhancers
    variables = levels.find_setting(sources, "variable_enhancers")
    for var in variables:
        for func in sources[var]["variable_enhancers"]:
            if isinstance(func, str):
                func = func_from_string(func)
            dss2 = func(dss2, var, folder)

    # TODO this info needs to be moved to `sources`.
    vars_for_sharpening = ['nmdi', 'bsi', 'mndwi', 'cos_solar_zangle', 'vari_red_edge', 'psri', 'nir', 'green']
    for x in vars_for_sharpening:
        if x in dss2.keys():
            _ = dss2.pop(x)

    # Align all the variables together.
    example_source = levels.find_setting(sources, "is_example", min_length = 1)
    if len(example_source) == 1:
        example_ds = dss[example_source[0]]
        log_example_ds(example_ds)
    elif len(example_source) == 0:
        example_ds = None
    else:
        log.warning(f"--> Multiple example datasets found, selecting lowest resolution.")
        example_ds = None

    # Open unopened netcdf files.
    dss2 = {**{k: open_ds(v) for k, v in dss2.items() if isinstance(v, str)}, 
            **{k:v for k,v in dss2.items() if not isinstance(v, str)}}

    spatial_interps = [sources[list(x.data_vars)[0]]["spatial_interp"] for x in dss2.values()]
    dss3, temp_files3 = align_pixels(dss2.values(), folder, spatial_interps, example_ds, fn_append = "_step2")
    cleanup.append(temp_files3)

    # Merge everything.
    ds = xr.merge(dss3)

    # Apply general enhancers.
    for func in general_enhancers:
        ds, label = apply_enhancer(ds, None, func)

    while os.path.isfile(final_path):
        final_path = final_path.replace(".nc", "_.nc")

    if ds.time.size == 0:
        log.warning("--> No valid data created (ds.time.size == 0).")
        return ds

    ds = save_ds(ds, final_path, encoding = "initiate",
                    label = f"Creating merged file `{os.path.split(final_path)[-1]}`.")

    for nc in list(chain.from_iterable(cleanup)):
        remove_ds(nc)

    return ds

# if __name__ == "__main__":

#     import numpy as np
#     import glob

#     dss = {
#         ("SENTINEL2", "S2MSI2A"): r"/Users/hmcoerver/Local/test_data/SENTINEL2/S2MSI2A.nc",
#         ("SENTINEL3", "SL_2_LST___"): r"/Users/hmcoerver/Local/test_data/SENTINEL3/SL_2_LST___.nc",
#         # ("VIIRSL1", "VNP02IMG"): r"/Users/hmcoerver/Local/test_data/VIIRSL1/VNP02IMG.nc",
#     }
#     example_source = ("SENTINEL2", "S2MSI2A")

#     sources = {
#             "ndvi": {"spatial_interp": "nearest", "temporal_interp": "linear"},
#             "lst": {"spatial_interp": "nearest", "temporal_interp": "linear"},
#             "bt": {"spatial_interp": "nearest", "temporal_interp": "linear"},
#                 }

#     # example_source = ("source1", "productX")
#     folder = r"/Users/hmcoerver/Local/test_data"
#     enhancers = list()
#     example_t_vars = ["lst"]

#     chunks = (1, 500, 500)

#     for fp in glob.glob(os.path.join(folder, "*.nc")):
#         os.remove(fp)

#     from pywapor.general.logger import log, adjust_logger
#     adjust_logger(True, folder, "INFO")

#     ds = main(dss, sources, example_source, folder, enhancers, example_t_vars = ["lst"])


import os
import datetime
import copy
import pandas as pd
import numpy as np
from osgeo import gdal
import pywapor.collect.protocol.cog as cog
from pywapor.collect.protocol.projections import get_crss
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.general.processing_functions import open_ds, process_ds, save_ds, remove_ds

def default_vars(product_name, req_vars = ["p"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["p"].

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    
    variables =  {
        "P05": {
            "precip": [("time", "lat", "lon"), "p"],
            "crs": [(), "spatial_ref"],
        }
    }

    req_dl_vars = {
        "P05": {
            "p": ["precip", "crs"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = ["p"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["p"].

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "P05": {
            "p": [],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name = "P05", req_vars = ["p"],
                variables = None, post_processors = None):
    """Download CHIRPS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str, optional
        Name of the product to download, by default "P05".
    req_vars : list, optional
        Which variables to download for the selected product, by default ["p"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "CHIRPS")

    if not os.path.isdir(folder):
        os.makedirs(folder)

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.05, latlim[1] + 0.05]
        lonlim = [lonlim[0] - 0.05, lonlim[1] + 0.05]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim], "t": ["time", timelim]}
    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    timedelta = pd.Timedelta(hours=12)
    data_source_crs = get_crss("WGS84")


    chirps_domain = [
        -180.0000000000000000,
        -50.0000014901161194, 
        180.0000053644180298,
        50.0000000000000000
        ]
    chirps_res = [7200, 2000]
    chirps_px_size = [(chirps_domain[2] - chirps_domain[0])/chirps_res[0],
                      (chirps_domain[3] - chirps_domain[1])/chirps_res[1]]
    
    def _snap_point(x0, xres, X):
        n = np.round((X - x0) / xres)
        Xsnap = n*xres + x0
        return Xsnap, n
    
    latlim_ = [
        _snap_point(chirps_domain[0], chirps_px_size[0], latlim[0])[0],
        _snap_point(chirps_domain[0], chirps_px_size[0], latlim[1])[0],
              ]
    lonlim_ = [
        _snap_point(chirps_domain[1], chirps_px_size[1], lonlim[0])[0],
        _snap_point(chirps_domain[1], chirps_px_size[1], lonlim[1])[0],
              ]

    warp_kwargs = dict()
    warp_kwargs["outputBounds"] = [lonlim_[0], latlim_[0], lonlim_[1], latlim_[1]]
    warp_kwargs["outputBoundsSRS"] = "epsg:4326"

    dates = pd.date_range(timelim[0], timelim[1], freq = "D")
    urls = [f"/vsicurl/https://data.chc.ucsb.edu/products/CHIRPS-2.0/global_daily/cogs/p05/{x.year}/chirps-v2.0.{x.year}.{x.month:>02}.{x.day:>02}.cog" for x in dates]
    temp_fn = fn.replace(".nc", "_temp.nc")
    temp_vrt = temp_fn.replace(".nc", ".vrt")
    
    gdal_config_options = {
        "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
    }
    try:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, v)
        x = cog.cog_dl(
                        urls, 
                        temp_fn, 
                       warp_kwargs=warp_kwargs
                       )
    except Exception as e:
        raise e
    finally:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, None)
    
    ds_ = open_ds(x)
    ds = ds_.to_array("time").to_dataset(name = "precip")
    ds = ds.assign_coords({"time": [np.datetime64(x + timedelta, "ns") for x in dates]})
    ds = process_ds(ds, coords, variables, crs = data_source_crs)

    ds = apply_enhancers(post_processors, ds)

    ds = save_ds(ds, fn, label = f"Merging files.")

    remove_ds(temp_fn)
    if os.path.isfile(temp_vrt):
        remove_ds(temp_vrt)

    return ds[req_vars_orig]

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/pywapor_test_chirps"
    # # latlim = [26.9, 33.7]
    # # lonlim = [25.2, 37.2]
    bb = [2.1737, 36.2283, 2.1916, 36.2390] # [xmin, ymin, xmax, ymax] #Wad_Helal
    timelim = [np.datetime64("2022-02-26"), np.datetime64("2022-02-26")] 
    latlim = bb[1::2]
    lonlim = bb[0::2]

    product_name = "P05"
    req_vars = ["p"]
    # CHIRPS.
    ds = download(folder, latlim, lonlim, timelim)



"""
@author: Radoslaw Guzinski
Copyright: (C) 2017, Radoslaw Guzinski
"""

import math
import os
import numpy as np
import scipy.ndimage as ndi
from numba import njit, stencil
from osgeo import gdal
from pyproj import Proj, Transformer
from pywapor.general.logger import log

def openRaster(raster):
    closeOnExit = False
    try:
        raster.GetProjection()
        openRaster = raster
    except AttributeError:
        openRaster = gdal.Open(str(raster))
        closeOnExit = True
    return openRaster, closeOnExit


def getRasterInfo(raster):
    r, closeOnExit = openRaster(raster)
    proj = r.GetProjection()
    gt = r.GetGeoTransform()
    sizeX = r.RasterXSize
    sizeY = r.RasterYSize
    extent = [gt[0], gt[3]+gt[5]*sizeY, gt[0]+gt[1]*sizeX, gt[3]]
    bands = r.RasterCount
    if closeOnExit:
        r = None
    return proj, gt, sizeX, sizeY, extent, bands


def resampleWithGdalWarp(srcFile, templateFile, outFile="", outFormat="MEM",
                         resampleAlg="average"):
    # Get template projection, extent and resolution
    proj, gt, sizeX, sizeY, extent, _ = getRasterInfo(templateFile)

    # Resample with GDAL warp
    outDs = gdal.Warp(str(outFile),
                      openRaster(srcFile)[0],
                      format=outFormat,
                      dstSRS=proj,
                      xRes=gt[1],
                      yRes=gt[5],
                      outputBounds=extent,
                      resampleAlg=resampleAlg)

    return outDs


def point2pix(point, gt, upperBound=False):
    mx = point[0]
    my = point[1]
    if not upperBound:
        px = math.floor((mx - gt[0]) / gt[1])  # x pixel
        py = math.floor((my - gt[3]) / gt[5])  # y pixel
    else:
        px = math.ceil((mx - gt[0]) / gt[1])  # x pixel
        py = math.ceil((my - gt[3]) / gt[5])  # y pixel
    return [int(px), int(py)]


def pix2point(pix, gt):
    px = pix[0]
    py = pix[1]
    mx = px*gt[1] + gt[0]  # x coordinate
    my = py*gt[5] + gt[3]  # y coordinate
    return [mx, my]


# save the data to geotiff or memory
def saveImg(data, geotransform, proj, outPath, noDataValue=None, fieldNames=[]):
    outPath = str(outPath)

    # Save to memory first
    is_netCDF = False
    memDriver = gdal.GetDriverByName("MEM")
    shape = data.shape
    if len(shape) > 2:
        ds = memDriver.Create("MEM", shape[1], shape[0], shape[2], gdal.GDT_Float32)
        ds.SetProjection(proj)
        ds.SetGeoTransform(geotransform)
        for i in range(shape[2]):
            ds.GetRasterBand(i+1).WriteArray(data[:, :, i])
    else:
        ds = memDriver.Create("MEM", shape[1], shape[0], 1, gdal.GDT_Float32)
        ds.SetProjection(proj)
        ds.SetGeoTransform(geotransform)
        ds.GetRasterBand(1).WriteArray(data)

    # Save to file if required
    if outPath == "MEM":
        if noDataValue is None:
            noDataValue = np.nan
        ds.GetRasterBand(1).SetNoDataValue(noDataValue)
    else:
        # If the output file has .nc extension then save it as netCDF,
        # otherwise assume that the output should be a GeoTIFF (COG)
        ext = os.path.splitext(outPath)[1]
        if ext.lower() == ".nc":
            fileFormat = "netCDF"
            driverOpt = ["FORMAT=NC2"]
            is_netCDF = True
        else:
            fileFormat = "COG"
            driverOpt = ['COMPRESS=DEFLATE', 'PREDICTOR=YES', 'BIGTIFF=IF_SAFER']
        out_ds = gdal.Translate(outPath, ds, format=fileFormat, creationOptions=driverOpt,
                                noData=noDataValue, stats=True)
        # If GDAL driers for other formats do not exist then default to GeoTiff
        if out_ds is None:
            log.warning("--> Selected GDAL driver is not supported! Saving as GeoTiff!")
            fileFormat = "GTiff"
            driverOpt = ['COMPRESS=DEFLATE', 'PREDICTOR=1', 'BIGTIFF=IF_SAFER']
            is_netCDF = False
            ds = gdal.Translate(outPath, ds, format=fileFormat, creationOptions=driverOpt,
                                noData=noDataValue, stats=True)
        else:
            ds = out_ds

        # In case of netCDF format use netCDF4 module to assign proper names
        # to variables (GDAL can't do this). Also it seems that GDAL has
        # problems assigning projection to all the bands so fix that.
        if is_netCDF and fieldNames:
            from netCDF4 import Dataset
            ds = None
            ds = Dataset(outPath, 'a')
            grid_mapping = ds["Band1"].grid_mapping
            for i, field in enumerate(fieldNames):
                ds.renameVariable("Band"+str(i+1), field)
                ds[field].grid_mapping = grid_mapping
            ds.close()
            ds = gdal.Open('NETCDF:"'+outPath+'":'+fieldNames[0])

    return ds


def binomialSmoother(data):
    def filterFunction(footprint):
        weight = [1, 2, 1, 2, 4, 2, 1, 2, 1]
        # Don't smooth land and invalid pixels
        if np.isnan(footprint[4]):
            return footprint[4]

        footprintSum = 0
        weightSum = 0
        for i in range(len(weight)):
            # Don't use land and invalid pixels in smoothing of other pixels
            if not np.isnan(footprint[i]):
                footprintSum = footprintSum + weight[i] * footprint[i]
                weightSum = weightSum + weight[i]
        try:
            ans = footprintSum/weightSum
        except ZeroDivisionError:
            ans = footprint[4]
        return ans

    smoothedData = ndi.filters.generic_filter(data, filterFunction, 3)

    return smoothedData


def appendNpArray(array, data, axis=None):
    if array is None or array.size == 0:
        array = data
    else:
        array = np.append(array, data, axis=axis)
    return array


# Reproject and subset the given low resolution datasets to high resolution
# scene projection and extent
def reprojectSubsetLowResScene(highResScene, lowResScene, resampleAlg=gdal.GRA_Bilinear):

    # Read the required metadata
    proj_HR, gt_HR, xsize_HR, ysize_HR, extent = getRasterInfo(highResScene)[0:5]
    proj_LR, gt_LR, xsize_LR, ysize_LR = getRasterInfo(lowResScene)[0:4]

    # Transform "middle pixel" and "middle pixel + 1" between LR and HR projections
    # to obtain LR resolution in HR projection. This method can handle different
    # x and y resolution
    midPix = [int(xsize_LR/2), int(ysize_LR/2)]
    midPix_2 = [midPix[0]+1, midPix[1]+1]
    midPoint = pix2point(midPix, gt_LR)
    midPoint_2 = pix2point(midPix_2, gt_LR)
    transformer = Transformer.from_proj(Proj(proj_LR), Proj(proj_HR), always_xy=True)
    x1, y1 = transformer.transform(midPoint[0], midPoint[1])
    x2, y2 = transformer.transform(midPoint_2[0], midPoint_2[1])
    xRes_LR_proj = x2 - x1
    yRes_LR_proj = y2 - y1
    # Now do the same with UL pixel to obtain low resolution geotransform in
    # the new projection
    UL_x, UL_y = transformer.transform(gt_LR[0], gt_LR[3])
    gt_LR = [UL_x, xRes_LR_proj, 0, UL_y, 0, yRes_LR_proj]

    # Now subset to high resolution scene extent while not shifting pixels
    UL = pix2point(point2pix([extent[0], extent[3]], gt_LR, upperBound=False), gt_LR)
    BR = pix2point(point2pix([extent[2], extent[1]], gt_LR, upperBound=True), gt_LR)
    out = gdal.Warp("",
                    openRaster(lowResScene)[0],
                    format="MEM",
                    dstSRS=proj_HR,
                    resampleAlg=gdal.GRA_NearestNeighbour,
                    xRes=gt_LR[1],
                    yRes=gt_LR[5],
                    outputBounds=[UL[0], BR[1], BR[0], UL[1]])

    return out


# Resample high res scene to low res pixel while extracting homogeneity
# statistics. It is assumed that both scenes have the same projection and extent.
def resampleHighResToLowRes(highResScene, lowResScene):

    gt_HR = getRasterInfo(highResScene)[1]
    bands_HR = getRasterInfo(highResScene)[5]
    gt_LR, xSize_LR, ySize_LR = getRasterInfo(lowResScene)[1:4]
    xRes_HR = gt_HR[1]
    yRes_HR = abs(gt_HR[5])
    xRes_LR = gt_LR[1]
    yRes_LR = gt_LR[5]

    aggregatedMean = np.zeros((ySize_LR,
                               xSize_LR,
                               bands_HR))
    aggregatedStd = np.zeros(aggregatedMean.shape)

    # Go through all the high res bands and calculate mean and standard
    # deviation when aggregated to the low resolution
    highRes, close = openRaster(highResScene)
    for band in range(bands_HR):
        bandData_HR = highRes.GetRasterBand(band+1).ReadAsArray()
        aggregatedMean[:, :, band], aggregatedStd[:, :, band] =\
            _resampleHighResToLowRes(bandData_HR, ySize_LR, yRes_LR, yRes_HR, xSize_LR, xRes_LR,
                                     xRes_HR, gt_HR, gt_LR)
    if close:
        highRes = None
    return aggregatedMean, aggregatedStd


@njit
def _resampleHighResToLowRes(bandData_HR, ySize_LR, yRes_LR, yRes_HR, xSize_LR, xRes_LR, xRes_HR,
                             gt_HR, gt_LR):
    aggregatedMean = np.zeros((ySize_LR, xSize_LR))
    aggregatedStd = np.zeros((ySize_LR, xSize_LR))
    for yPix_LR in range(ySize_LR):
        yPos_LR_min = gt_LR[3] + yPix_LR*yRes_LR
        yPix_HR_min = int(round(max(0, gt_HR[3] - yPos_LR_min) / yRes_HR))
        yPix_HR_max = int(round(max(0, gt_HR[3] - (yPos_LR_min + yRes_LR)) / yRes_HR))
        for xPix_LR in range(xSize_LR):
            xPos_LR_min = gt_LR[0] + xPix_LR*xRes_LR
            xPix_HR_min = int(round(max(0, xPos_LR_min - gt_HR[0]) / xRes_HR))
            xPix_HR_max = int(round(max(0, xPos_LR_min + xRes_LR - gt_HR[0]) / xRes_HR))
            aggregatedMean[yPix_LR, xPix_LR] =\
                np.nanmean(bandData_HR[yPix_HR_min:yPix_HR_max, xPix_HR_min:xPix_HR_max])
            aggregatedStd[yPix_LR, xPix_LR] =\
                np.nanstd(bandData_HR[yPix_HR_min:yPix_HR_max, xPix_HR_min:xPix_HR_max])

    return aggregatedMean, aggregatedStd


@stencil(cval=1.0)
def removeEdgeNaNs(a):
    if np.isnan(a[0, 0]) and (not np.isnan(a[-1, 0]) or not np.isnan(a[1, 0]) or
                              not np.isnan(a[0, -1]) or not np.isnan(a[0, 1])):
        return np.nanmean(np.array([a[-1, 0], a[1, 0], a[0, -1], a[0, 1]]))
    else:
        return a[0, 0]


__all__ = [
    "LANDSAT", 
    "CHIRPS", 
    "ERA5", 
    "GEOS5", 
    "GLOBCOVER", 
    "MERRA2", 
    "MODIS", 
    "TERRA", 
    "SRTM", 
    "STATICS", 
    "SENTINEL2", 
    "SENTINEL3", 
    "VIIRSL1", 
    "COPERNICUS", 
    "LSASAF"
    ]
from . import LANDSAT, CHIRPS, ERA5, GEOS5, GLOBCOVER, MERRA2, MODIS, TERRA, SRTM, STATICS, SENTINEL2, SENTINEL3, VIIRSL1, COPERNICUS, LSASAF
